//
// Example of using bitonic sort to sort 2D texture of RGBA values
//
// Author: Alex V. Boreskoff, <steps3d@narod.ru>
//
//

#include	"libExt.h"
#include	"libTexture.h"
#include	"fpTexture.h"
#include	"FrameBuffer.h"
#include	"GlslProgram.h"
#include	"utils.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>

#define	SIZE1	8
#define	SIZE2	8

int			totalPasses = 0;
bool		first = true;
GLuint		dataMap;
GLenum		mapFormat;
GlslProgram	progRow;
GlslProgram	progCol;
FrameBuffer	* buf [2];	
float	res [4*SIZE1*SIZE2];

void	doRowStage     ( int stage, int colSpan, float initSign, int& index );
void	doLastRowStage ( int stage, int colSpan, float initSign, int& index );

int	intLog ( int value )
{
	int	i = 0;

	for ( int n = value; n; n >>=1, i++ )
		;

	i--;

	if ( value != (1 << i) )
		printf ( "Not a power of 2.\n" );
		
	return i;
}

bool	isPowerOfTwo ( int n )
{
	return (n & (n-1)) == 0;
}

void	printArray ( const char * title, float * buf )
{
	printf ( "\na [%d,%d] =\n", SIZE1, SIZE2 );
	
	for ( int i = 0; i < SIZE1*SIZE2; i++ )
	{
		printf ( "%2.0f ", buf [4*i + 3] );
		
		if ( (i % SIZE1) == SIZE1-1 )
			printf ( "\n" );
	}

	printf ( "\n" );
}

void	createDataMap ()
{
	float	values [4*SIZE1*SIZE2];
	int		index = 0;

	for ( int i = 0; i < SIZE1; i++ )
		for ( int j = 0; j < SIZE2; j++ )
		{
			values [index++] = 0;
			values [index++] = 0;
			values [index++] = 0;
			values [index++] = rand () % 100;
		}

	printArray ( "Source", values );

    glGenTextures   ( 1, &dataMap );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, dataMap );
    glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );                         // set 1-byte alignment

    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP );    // set texture to repeat mode
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri ( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	
    glTexImage2D    ( GL_TEXTURE_RECTANGLE_ARB, 0, mapFormat, SIZE1, SIZE2, 0, GL_RGBA, GL_FLOAT, values );
    glBindTexture   ( GL_TEXTURE_RECTANGLE_ARB, 0 );
}

void	rowSort ( int colSpan, float initSign, int& index )
{
	int	stage;
	int	numStages = intLog ( SIZE1 );
									
	printf ( "rowSort\n" );

									// do stages with comparing order changing 
	for ( stage = 1; stage < numStages; stage++ )
		doRowStage ( stage, colSpan, initSign, index );
		
									// do last stage with a fixed order
	doLastRowStage ( stage, colSpan, initSign, index );
}

void	doRowStage ( int stage, int colSpan, float initSign, int& index )
{
	int	numPasses    = stage;
	int	blockSize    = 1 << stage;
	int	numBlocks    = SIZE1 / blockSize;
	int	step         = blockSize / 2;
	int	subBlockSize = blockSize;
	int	numSubBlocks = 1;
	
	for ( int pass = 1; pass <= numPasses; pass++, step >>= 1, index ^= 1, totalPasses++ )
	{
		int	x = 0;
		
		glActiveTextureARB ( GL_TEXTURE0_ARB );

		if ( first )
		{
			glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, dataMap );

			first = false;
		}
		else
			glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index] -> getColorBuffer () );
	
		buf [index ^ 1] -> bind ();
		progRow.bind ();
		progRow.setUniformFloat ( "step", step );
	
		startOrtho ( SIZE1, SIZE2 );

		glBegin ( GL_QUADS );			// now draw blocks
		
		for ( int i = 0; i < numBlocks; i++ )
		{
			float	sign = initSign * ((i & 1) == 0 ? 1 : -1);
			
			for ( int j = 0; j < numSubBlocks; j++, x += subBlockSize )
			{
										// pass tex coords - z - first/second pixel, 
										// w - ascending/descending
				float s = sign;
				
				for ( int y = 0; y < SIZE2; y += colSpan, s = -s )
				{
					glMultiTexCoord4f ( GL_TEXTURE0_ARB, x, y, -1, s );
					glVertex2f        ( x, y );
					
					glMultiTexCoord4f ( GL_TEXTURE0_ARB, x + subBlockSize, y, 1, s );
					glVertex2f        ( x + subBlockSize, y );

					glMultiTexCoord4f ( GL_TEXTURE0_ARB, x + subBlockSize, y + colSpan, 1, s  );
					glVertex2f        ( x + subBlockSize, y + colSpan );

					glMultiTexCoord4f ( GL_TEXTURE0_ARB, x, y + colSpan, -1, s  );
					glVertex2f        ( x, y + colSpan );
				}
			}
		}
		
		glEnd   ();
		glFlush ();
		
		endOrtho ();

		progRow.unbind ();

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index ^ 1] -> getColorBuffer () );
		glReadPixels  ( 0, 0, SIZE1, SIZE2, GL_RGBA, GL_FLOAT, res );

		buf [index ^ 1] -> unbind ();
		
		printf     ( "Row stage %3d, pass %3d\n", stage, pass );
		printArray ( "", res );

		numSubBlocks *= 2;
		subBlockSize /= 2;
	}
}

void	doLastRowStage ( int stage, int colSpan, float initSign, int& index )
{
	int	numPasses    = stage;
	int	blockSize    = 1 << stage;
	int	step         = blockSize / 2;
	int	subBlockSize = blockSize;
	int	numSubBlocks = 1;
	
	for ( int pass = 1; pass <= numPasses; pass++, step >>= 1, index ^= 1, totalPasses++ )
	{
		int	x = 0;
		
		glActiveTextureARB ( GL_TEXTURE0_ARB );

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index] -> getColorBuffer () );
	
		buf [index ^ 1] -> bind ();
		progRow.bind ();
		progRow.setUniformFloat ( "step", step );
	
		startOrtho ( SIZE1, SIZE2 );

										// now draw blocks
		glBegin ( GL_QUADS );
		
		for ( int j = 0; j < numSubBlocks; j++, x += subBlockSize )
		{
										// pass tex coords - z - first/second pixel, 
										// w - ascending/descending
			float	s = initSign;
			
			for ( int y = 0; y < SIZE2; y += colSpan, s = -s )
			{
				glMultiTexCoord4f ( GL_TEXTURE0_ARB, x, y, -1, s );
				glVertex2f        ( x, y );
				
				glMultiTexCoord4f ( GL_TEXTURE0_ARB, x + subBlockSize, y, 1, s );
				glVertex2f        ( x + subBlockSize, y );

				glMultiTexCoord4f ( GL_TEXTURE0_ARB, x + subBlockSize, y + colSpan, 1, s  );
				glVertex2f        ( x + subBlockSize, y + colSpan );

				glMultiTexCoord4f ( GL_TEXTURE0_ARB, x, y + colSpan, -1, s  );
				glVertex2f        ( x, y + colSpan );
			}
		}
		
		glEnd   ();
		glFlush ();
		
		endOrtho ();

		progRow.unbind ();

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index ^ 1] -> getColorBuffer () );
		glReadPixels  ( 0, 0, SIZE1, SIZE2, GL_RGBA, GL_FLOAT, res );

		buf [index ^ 1] -> unbind ();
		
		printf     ( "Last row stage %3d, pass %3d\n", stage, pass );
		printArray ( "", res );

		numSubBlocks *= 2;
		subBlockSize /= 2;
	}
}

void	doColumnStage ( int stage, int& index )
{
	int	numPasses    = stage;
	int	blockSize    = 1 << stage;
	int	numBlocks    = SIZE2 / blockSize;
	int	step         = blockSize / 2;
	int	subBlockSize = blockSize;
	int	numSubBlocks = 1;
	
	for ( int pass = 1; pass <= numPasses; pass++, step >>= 1, totalPasses++ )
	{
		int	y = 0;
		
		glActiveTextureARB ( GL_TEXTURE0_ARB );

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index] -> getColorBuffer () );
	
		buf [index ^ 1] -> bind ();
		progCol.bind ();
		progCol.setUniformFloat ( "step", step );
	
		startOrtho ( SIZE1, SIZE2 );

		glBegin ( GL_QUADS );			// now draw blocks
		
		for ( int i = 0; i < numBlocks; i++ )
		{
			float	sign = (i & 1) == 0 ? 1 : -1;
			
			for ( int j = 0; j < numSubBlocks; j++, y += subBlockSize )
			{
										// pass tex coords - z - first/second pixel, 
										// w - ascending/descending
				glMultiTexCoord4f ( GL_TEXTURE0_ARB, 0, y, -1, sign );
				glVertex2f        ( 0, y );
				
				glMultiTexCoord4f ( GL_TEXTURE0_ARB, 0, y + subBlockSize, 1, sign );
				glVertex2f        ( 0, y + subBlockSize );

				glMultiTexCoord4f ( GL_TEXTURE0_ARB, SIZE1, y + subBlockSize, 1, sign  );
				glVertex2f        ( SIZE1, y + subBlockSize );

				glMultiTexCoord4f ( GL_TEXTURE0_ARB, SIZE1, y, -1, sign  );
				glVertex2f        ( SIZE1, y );
			}
		}
		
		glEnd   ();
		glFlush ();
		
		endOrtho ();

		progCol.unbind ();

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index ^ 1] -> getColorBuffer () );
		glReadPixels  ( 0, 0, SIZE1, SIZE2, GL_RGBA, GL_FLOAT, res );

		buf [index ^ 1] -> unbind ();
		
		printf     ( "Column stage %3d, pass %3d\n", stage, pass );
		printArray ( "", res );

		numSubBlocks *= 2;
		subBlockSize /= 2;
		index        ^= 1;
	}

	doLastRowStage ( intLog ( SIZE1 ), blockSize, 1, index );
}

void	doLastColumnStage ( int stage, int& index )
{
	int	numPasses    = stage;
	int	blockSize    = 1 << stage;
	int	step         = blockSize / 2;
	int	subBlockSize = blockSize;
	int	numSubBlocks = 1;
	
	for ( int pass = 1; pass <= numPasses; pass++, step >>= 1, totalPasses++ )
	{
		int	y = 0;
		
		glActiveTextureARB ( GL_TEXTURE0_ARB );

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index] -> getColorBuffer () );
	
		buf [index ^ 1] -> bind ();
		progCol.bind ();
		progCol.setUniformFloat ( "step", step );
	
		startOrtho ( SIZE1, SIZE2 );

		glBegin ( GL_QUADS );			// now draw blocks
		
		for ( int j = 0; j < numSubBlocks; j++, y += subBlockSize )
		{
										// pass tex coords - z - first/second pixel, 
										// w - ascending/descending
			glMultiTexCoord4f ( GL_TEXTURE0_ARB, 0, y, -1, 1 );
			glVertex2f        ( 0, y );
			
			glMultiTexCoord4f ( GL_TEXTURE0_ARB, 0, y + subBlockSize, 1, 1 );
			glVertex2f        ( 0, y + subBlockSize );

			glMultiTexCoord4f ( GL_TEXTURE0_ARB, SIZE1, y + subBlockSize, 1, 1  );
			glVertex2f        ( SIZE1, y + subBlockSize );

			glMultiTexCoord4f ( GL_TEXTURE0_ARB, SIZE1, y, -1, 1  );
			glVertex2f        ( SIZE1, y );
		}
		
		glEnd   ();
		glFlush ();
		
		endOrtho ();

		progCol.unbind ();

		glBindTexture ( GL_TEXTURE_RECTANGLE_ARB, buf [index ^ 1] -> getColorBuffer () );
		glReadPixels  ( 0, 0, SIZE1, SIZE2, GL_RGBA, GL_FLOAT, res );

		buf [index ^ 1] -> unbind ();
		
		printf     ( "Last column stage %3d, pass %3d\n", stage, pass );
		printArray ( "", res );

		numSubBlocks *= 2;
		subBlockSize /= 2;
		index        ^= 1;
	}

	doLastRowStage ( intLog ( SIZE1 ), blockSize, 1, index );
}

void	doBitonicSort ()
{
	startOrtho ( SIZE1, SIZE2 );
	
	int	numStages, stage;
	int	index = 0;
	
	glEnable ( GL_TEXTURE_RECTANGLE_ARB );
									
	printf ( "Row sort\n" );
	
	rowSort ( 1, 1, index );	
								
	printf ( "Column sort\n" );
	
	numStages = intLog ( SIZE2 );
									
									// do stages with comparing order changing 
	for ( stage = 1; stage < numStages; stage++ )
		doColumnStage ( stage, index );
		
									// do last stage with a fixed order
	doLastColumnStage ( stage, index );
	
	endOrtho ();

	printf ( "-------------------------\nTotal passes: %d\n", totalPasses );
}

void	display ()
{
	createDataMap ();
	doBitonicSort ();
	
	exit ( 1 );
}

void	reshape ( int w, int h )
{
	glViewport ( 0, 0, w, h );
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 512, 512 );

								// create window
	glutCreateWindow ( "GPU Sorting 2D" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );

	init                      ();
	initExtensions            ();
	assertExtensionsSupported ( "ARB_texture_rectangle" );
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}

	mapFormat = fpFormatWithPrecision ( 32 );
	
	if ( mapFormat == 0 )
		return 1;
		
	for ( int i = 0; i < 2; i++ )
	{
		buf [i] = new FrameBuffer ( SIZE1, SIZE2 );
		
		unsigned tempMap = buf [i] -> createColorRectTexture ( GL_RGBA, mapFormat );
		
		buf [i] -> create ();
		buf [i] -> bind   ();
		
		if ( !buf [i] -> attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, tempMap ) )
			printf ( "buffer error with color attachment\n");

		if ( !buf [i] -> isOk () )
			printf ( "Error with framebuffer\n" );
						
		buf [i] -> unbind ();
	}
		
    if ( !progRow.loadShaders ( "sort2.vsh", "sort2-row.fsh" ) )
    {
        printf ( "Error loading sort1 shaders:\n%s\n", progRow.getLog ().c_str () );

        return 3;
    }
		
    if ( !progCol.loadShaders ( "sort2.vsh", "sort2-col.fsh" ) )
    {
        printf ( "Error loading sort1 shaders:\n%s\n", progCol.getLog ().c_str () );

        return 3;
    }
		
	glutMainLoop ();

	return 0;
}
